"""admin commands for cmem command line interface."""
import click

import jwt

from cmem.cmemc.cli import completion
from cmem.cmemc.cli.commands.user import user
from cmem.cmemc.cli.context import ApplicationContext
from cmem.cmemc.cli.utils import struct_to_table
from cmem.cmempy.api import (
    get_access_token, get_token
)
from cmem.cmemc.cli.commands import CmemcCommand, CmemcGroup
from cmem.cmemc.cli.commands.metrics import metrics
from cmem.cmemc.cli.commands.store import store
from cmem.cmemc.cli.commands.workspace import workspace
from cmem.cmempy.health import get_complete_status_info


@click.command(cls=CmemcCommand, name="status")
@click.option(
    "--key", "key",
    autocompletion=completion.status_keys,
    help="Get only specific key(s) from the status / info output. There are "
         "two special keys available: 'all' will list all available keys in "
         "the table, 'overall.healthy' with result in  UP in case all "
         "health flags are UP as well (otherwise DOWN)."
)
@click.option(
    "--exit-1",
    type=click.Choice(["never", "error", "always"]),
    default="never",
    show_default=True,
    help="Specify, when this command returns with exit code 1. Available options are "
         "'never' (exit 0 on errors and warnings), "
         "'error' (exit 1 on errors, exit 0 on warnings), "
         "'always' (exit 1 on errors and warnings)."
)
@click.option(
    "--enforce-table",
    is_flag=True,
    help="A single value with --key will be returned as plain text instead "
         "of a table with one row and the header. This default behaviour "
         "allows for more easy integration with scripts. This flag enforces "
         "the use of tabular output, even for single row tables."
)
@click.option(
    "--raw",
    is_flag=True,
    help="Outputs combined raw JSON output of the health/info endpoints."
)
@click.pass_obj
def status_command(app: ApplicationContext, key, exit_1, enforce_table, raw):
    """Output health and version information.

    This command outputs version and health information of the
    selected deployment. If the version information cannot be retrieved,
    UNKNOWN is shown.

    Additionally, this command warns you if the target version of your
    cmemc client is newer than the version of your backend and if the
    ShapeCatalog has a different version than your DataPlatform
    component.

    To get status information of all configured
    deployments use this command in combination with parallel.

    Example: cmemc config list | parallel --ctag cmemc -c {} admin status
    """
    _ = get_complete_status_info()
    if "error" in _["di"]:
        app.echo_debug(_["di"]["error"])
    if "error" in _["dp"]:
        app.echo_debug(_["dp"]["error"])
    if "error" in _["dm"]:
        app.echo_debug(_["dm"]["error"])
    basic_status = (
        _["dp"]["healthy"], _["di"]["healthy"], _["dm"]["healthy"],
        _["shapes"]["healthy"], _["store"]["healthy"]
    )
    if exit_1 in ("always", "error")\
            and ("DOWN" in basic_status or "UNKNOWN" in basic_status):
        raise ValueError(
            f"One or more major status flags are DOWN or UNKNOWN: {repr(_)}",
        )
    if raw:
        app.echo_info_json(_)
        return
    if key:
        table = [
            line for line
            in struct_to_table(_)
            if line[0].startswith(key) or key == "all"
        ]
        if len(table) == 1 and not enforce_table:
            app.echo_info(table[0][1])
            return
        if len(table) == 0:
            raise ValueError(
                f"No values for key(s): {key}"
            )
        app.echo_info_table(
            table,
            headers=["Key", "Value"],
            sort_column=0
        )
        return
    table = [
        ("DP", _["dp"]["version"], _["dp"]["healthy"]),
        ("DI", _["di"]["version"], _["di"]["healthy"]),
        ("DM", _["dm"]["version"], _["dm"]["healthy"]),
        ("SHAPES", _["shapes"]["version"], _["shapes"]["healthy"]),
        (_["store"]["type"], _["store"]["version"], _["store"]["healthy"])
    ]
    app.echo_info_table(
        table,
        headers=["Component", "Version", "Status"],
        sort_column=0
    )
    app.check_versions()
    if _["shapes"]["version"] not in (_["dp"]["version"], "UNKNOWN"):
        output = "Your ShapeCatalog version does not match your DataPlatform "\
                 "version. Please consider updating your bootstrap data."
        if exit_1 == "always":
            raise ValueError(output)
        app.echo_warning(output)


@click.command(cls=CmemcCommand, name="token")
@click.option(
    "--raw",
    is_flag=True,
    help="Outputs raw JSON. Note that this option will always try to fetch "
         "a new JSON token response. In case you are working with "
         "OAUTH_GRANT_TYPE=prefetched_token, this may lead to an error."
)
@click.option(
    "--decode",
    is_flag=True,
    help="Decode the access token and outputs the raw JSON. Note that the "
         "access token is only decoded and esp. not validated."
)
@click.pass_obj
def token_command(app, raw, decode):
    """Fetch and output an access token.

    This command can be used to check for correct authentication as well as
    to use the token with wget / curl or similar standard tools:

    Example: curl -H "Authorization: Bearer $(cmemc -c my admin token)"
    $(cmemc -c my config get DP_API_ENDPOINT)/api/custom/slug

    Please be aware that this command can reveal secrets which you might
    not want to be present in log files or on the screen.
    """
    # Note:
    # - get_access_token returns the token string which is maybe from conf
    # - get_token fetches a new token incl. envelope from keycloak

    if decode:
        token = get_access_token()
        decoded_token = jwt.decode(token, options={"verify_signature": False})
        if raw:
            app.echo_info_json(decoded_token)
        else:
            table = struct_to_table(decoded_token)
            app.echo_info_table(
                table,
                headers=["Key", "Value"],
                sort_column=0
            )
    else:
        if raw:
            app.echo_info_json(get_token())
        else:
            token = get_access_token()
            app.echo_info(token)


@click.group(cls=CmemcGroup)
def admin():
    """Import bootstrap data, backup/restore workspace or get status.

    This command group consists of commands for setting up and
    configuring eccenca Corporate Memory.
    """


admin.add_command(status_command)
admin.add_command(token_command)
admin.add_command(metrics)
admin.add_command(workspace)
admin.add_command(store)
admin.add_command(user)



"""Generate a single page documentation Markdown."""
from click import Command, Group


def print_manual(ctx):
    """Output the complete manual.

    Returns: None
    """
    print("""# Command Reference

This section lists the help texts of all commands as a reference
and to search for it.
""")
    print_group_manual_recursive(ctx.command, ctx=ctx)


def print_group_manual_recursive(command_group, ctx=None, prefix=""):
    """Output the help text of a command group (recursive)."""
    commands = command_group.commands
    for key in commands:
        command = commands[key]
        formatter = ctx.make_formatter()
        if isinstance(command, Group):
            print(f"## Command group: {prefix}{key}\n")
            print("```")
            command.format_help(ctx, formatter)
            print(formatter.getvalue().rstrip("\n"))
            print("```\n")

            print_group_manual_recursive(
                command,
                ctx=ctx,
                prefix=f"{prefix}{key} "
            )
        elif isinstance(command, Command):
            print(f"### Command: {prefix}{key}\n")
            print("```text")
            command.format_help(ctx, formatter)
            print(formatter.getvalue().rstrip("\n"))
            print("```\n")
        else:
            pass



"""Utility functions for CLI interface."""
import json
import os
import re
import unicodedata
from typing import Dict

from importlib.metadata import version
from prometheus_client import Metric
from prometheus_client.parser import text_string_to_metric_families

from cmem.cmempy.dp.admin import get_prometheus_data
from cmem.cmempy.dp.proxy.graph import get_graphs_list

NAMESPACES = {
    "void": "http://rdfs.org/ns/void#",
    "di": "https://vocab.eccenca.com/di/",
    "shui": "https://vocab.eccenca.com/shui/",
    "dsm": "https://vocab.eccenca.com/dsm/",
    "owl": "http://www.w3.org/2002/07/owl#",
    "sd": "http://www.w3.org/ns/sparql-service-description#"
}


def get_version():
    """Get the current version or SNAPSHOT."""
    # pylint: disable=import-outside-toplevel
    return version('cmem-cmemc')


def extract_error_message(error):
    """Extract a message from an exception."""
    # exceptions with response is HTTPError
    message = type(error).__name__ + ": " + str(error) + "\n"
    try:
        # try to load Problem Details for HTTP API JSON
        details = json.loads(error.response.text)
        message += type(error).__name__ + ": "
        if 'title' in details:
            message += details["title"] + ": "
        if 'detail' in details:
            message += details["detail"]
    except (AttributeError, ValueError):
        # is not json or any other issue, output plain response text
        pass
    return message


def is_completing():
    """Test for environment which indicates that we are in completion mode.

    Returns true if in validation mode, otherwise false.

    Returns: boolean
    """
    comp_words = os.getenv("COMP_WORDS", default=None)
    cmemc_complete = os.getenv("_CMEMC_COMPLETE", default=None)
    if comp_words is not None:
        return True
    if cmemc_complete is not None:
        return True
    return False


def iri_to_qname(iri):
    """Return a qname for an IRI based on well known namespaces.

    In case of no matching namespace, the full IRI is returned.

    Args:
        iri:

    Returns: string
    """
    for prefix, namespace in NAMESPACES.items():
        iri = iri.replace(namespace, prefix + ":")
    return iri


def read_rdf_graph_files(directory_path):
    """Read all files from directory_path and output as tuples.

    The tuple format is (filepath, graph_name),
    for example ("/tmp/rdf.nt", "http://example.com")
    """
    rdf_graphs = []
    for root, _, files in os.walk(directory_path):
        for _file in files:
            full_file_path = os.path.join(root, _file)
            graph_file_name = _file + ".graph"
            full_graph_file_name_path = os.path.join(
                root,
                graph_file_name
            )
            if (not _file.endswith(".graph")
               and os.path.exists(full_graph_file_name_path)):
                graph_name = read_file_to_string(
                    full_graph_file_name_path
                ).strip()
                rdf_graphs.append(
                    (full_file_path, graph_name)
                )
    return rdf_graphs


def read_file_to_string(file_path):
    """Read file to string."""
    with open(file_path, "rb") as _file:
        return _file.read().decode("utf-8")


def get_graphs(writeable=True, readonly=True):
    """Retrieve list of accessible graphs from DP endpoint.

    readonly=True|writeable=True outputs all graphs
    readonly=False|writeable=True outputs only writeable graphs
    readonly=True|writeable=False outputs graphs without write access
    (but read access)
    """
    all_graphs = get_graphs_list()
    filtered_graphs = []
    for graph in all_graphs:
        if graph['writeable'] and writeable:
            filtered_graphs.append(graph)
        if not graph['writeable'] and readonly:
            filtered_graphs.append(graph)
    return filtered_graphs


def get_graphs_as_dict(writeable=True, readonly=True):
    """Get the graph response as dict with IRI as main key."""
    graph_dict = {}
    for graph in get_graphs(writeable=writeable, readonly=readonly):
        graph_dict[graph["iri"]] = graph
    return graph_dict


def convert_uri_to_filename(value, allow_unicode=False):
    """Convert URI to unix friendly filename.

    Convert to ASCII if 'allow_unicode' is False. Convert spaces or repeated
    dashes to single dashes. Remove characters that aren't alphanumerics,
    underscores, or hyphens. Convert / to underscore. Convert to lowercase.
    Also strip leading and trailing whitespace, dashes, and underscores.
    """
    value = str(value)
    if allow_unicode:
        value = unicodedata.normalize('NFKC', value)
    else:
        value = unicodedata.normalize('NFKD', value)\
            .encode('ascii', 'ignore').decode('ascii')
    value = re.sub(r'\.', '_', value.lower())
    value = re.sub(r'/', '_', value.lower())
    value = re.sub(r'[^\w\s-]', '', value.lower())
    return re.sub(r'[-\s]+', '-', value).strip('-_')


def struct_to_table(source, table=None, prefix=""):
    """Prepare flat key/value table from a deep structure.

    This function takes structure and creates a flat table out of it,
    key by key, value by value. For each level deeper it prefixes the
    father keys.

    Example input:  {'k1': '1', 'k2': {'k3': '3', 'k4': '4'}}
    Example output: [['k1', '1'], ['k2:k3', '3'], ['k2:k4', '4']]

    Args:
        source (any): The structure which is transformed into a flat table.
        table (list): The table where the key/value rows will be appended.
        prefix (str): A prefix which is used to indicate the level.

    Returns:
        The input table extended with rows from the input source.
    """
    if table is None:
        table = []
    if type(source) in (str, bool, int, float):
        table.append([prefix, source])
        return table
    if isinstance(source, dict):
        if len(prefix) != 0:
            prefix = prefix + "."
        for key in source:
            table = struct_to_table(source[key], table, prefix + key)
        return table
    if isinstance(source, list):
        for value in source:
            table = struct_to_table(value, table, prefix)
        return table
    return table


def split_task_id(task_id):
    """Validate and split cmemc task ID.

    Args:
        task_id (str): The task ID in the workspace.

    Raises:
        ValueError: in case the task ID is not splittable
    """
    try:
        project_part = task_id.split(":")[0]
        task_part = task_id.split(":")[1]
    except IndexError as error:
        raise ValueError(
            f"{task_id} is not a valid task ID."
        ) from error
    return project_part, task_part


def metric_get_labels(family: Metric, clean=True) -> Dict[str, list[str]]:
    """
    Get the labels of a metric family.

    Args:
        family: the metric family
        clean : remove keys with only one dimension

    Returns: labels as dict
    """
    labels: Dict[str, list[str]] = {}
    # build tree structure
    for sample in family.samples:
        for label in sample.labels:
            value = sample.labels[label]
            if label not in labels:
                labels[label] = []
            if value not in labels[label]:
                labels[label].append(value)
    if clean:
        labels = dict(filter(lambda elem: len(elem[1]) > 1, labels.items()))
    return labels


def metrics_get_dict(job_id="DP"):
    """Get metrics data as dict."""
    data = {}
    if job_id == "DP":
        for family in text_string_to_metric_families(
                get_prometheus_data().text
        ):
            data[family.name] = family
        return data
    raise ValueError(
        f"job name {job_id} unknown."
    )


def metrics_get_family(job_id, metric_id):
    """Get family data.

    This function returns a dictionary of metric families.
    """
    try:
        return metrics_get_dict(job_id=job_id)[metric_id]
    except KeyError as error:
        raise ValueError(
            f"The job {job_id} does not have a metric {metric_id}."
        ) from error



"""Utility functions for CLI auto-completion functionality."""
# pylint: disable=unused-argument, broad-except, too-many-lines
import os
from collections import Counter
from contextlib import suppress

import requests
from bs4 import BeautifulSoup
from click import Context
from natsort import natsorted, ns
from prometheus_client.parser import text_string_to_metric_families

from cmem.cmemc.cli.context import CONTEXT
from cmem.cmemc.cli.utils import (
    get_graphs,
    metric_get_labels,
    metrics_get_dict, struct_to_table
)
from cmem.cmempy.dp.admin import get_prometheus_data
from cmem.cmempy.health import get_complete_status_info
from cmem.cmempy.keycloak.group import list_groups
from cmem.cmempy.keycloak.user import list_users, get_user_by_username, user_groups
from cmem.cmempy.plugins.marshalling import get_marshalling_plugins
from cmem.cmempy.queries import QUERY_CATALOG, get_query_status
from cmem.cmempy.vocabularies import get_vocabularies
from cmem.cmempy.workflow.workflows import get_workflows_io
from cmem.cmempy.workspace import (
    get_task_plugin_description,
    get_task_plugins,
)
from cmem.cmempy.workspace.projects.project import get_projects
from cmem.cmempy.workspace.projects.resources import get_all_resources
from cmem.cmempy.workspace.python import list_packages
from cmem.cmempy.workspace.search import list_items

SORT_BY_KEY = 0
SORT_BY_DESC = 1


def _finalize_completion(
        candidates: list,
        incomplete: str = '',
        sort_by: int = SORT_BY_KEY,
        nat_sort: bool = False,
        reverse: bool = False
) -> list:
    """Sort and filter candidates list.

    candidates are sorted with natsort library by sort_by key.

    Args:
        candidates (list):  completion dictionary to filter
        incomplete (str):   incomplete string at the cursor
        sort_by (str):      SORT_BY_KEY or SORT_BY_DESC
        nat_sort (bool):    if true, uses the natsort package for sorting
        reverse (bool):     if true, sorts in reverse order

    Returns:
        filtered and sorted candidates list

    Raises:
        ValueError in case of wrong sort_by parameter
    """
    if sort_by not in (SORT_BY_KEY, SORT_BY_DESC):
        raise ValueError("sort_by should be 0 or 1.")
    incomplete = incomplete.lower()
    if len(candidates) == 0:
        return candidates
    # remove duplicates
    candidates = list(set(candidates))

    if isinstance(candidates[0], str):
        # list of strings filtering and sorting
        filtered_candidates = [
            element for element in candidates
            if element.lower().find(incomplete) != -1
        ]
        if nat_sort:
            return natsorted(
                seq=filtered_candidates,
                alg=ns.IGNORECASE,
                reverse=reverse
            )
        # this solves that case-insensitive sorting is not stable in ordering
        # of "equal" keys (https://stackoverflow.com/a/57923460)
        return sorted(
            filtered_candidates,
            key=lambda x: (str(x).casefold(), x),
            reverse=reverse
        )
    if isinstance(candidates[0], tuple):
        # list of tuples filtering and sorting
        filtered_candidates = [
            element for element in candidates
            if str(element[0]).lower().find(incomplete) != -1
            or str(element[1]).lower().find(incomplete) != -1
        ]
        if nat_sort:
            return natsorted(
                seq=filtered_candidates,
                key=lambda k: k[sort_by],   # type: ignore
                alg=ns.IGNORECASE,
                reverse=reverse
            )
        return sorted(
            filtered_candidates,
            key=lambda x: (str(x[sort_by]).casefold(), str(x[sort_by])),
            reverse=reverse
        )
    raise ValueError(
        "candidates should be a list of strings or a list of tuples."
    )


def add_metadata_parameter(list_=None):
    """Extend a list with metadata keys and key descriptions."""
    if list_ is None:
        list_ = []
    list_.insert(
        0,
        ("description", "Metadata: A description.")
    )
    list_.insert(
        0,
        ("label", "Metadata: A name.")
    )
    return list_


def dataset_parameter(ctx, args, incomplete):
    """Prepare a list of dataset parameters for a dataset type."""
    CONTEXT.set_connection_from_args(args)
    incomplete = incomplete.lower()
    # look if cursor is in value position of the -p option and
    # return nothing in case it is (values are not completed atm)
    if args[len(args) - 2] in ("-p", "--parameter"):
        return []
    # try to determine the dataset type
    dataset_type = None
    for num, arg in enumerate(args):
        if arg == "--type":
            dataset_type = args[num + 1]
    # without type, we know nothing
    if dataset_type is None:
        return []
    plugin = get_task_plugin_description(dataset_type)
    properties = plugin["properties"]
    options = []
    for key in properties:
        title = properties[key]["title"]
        description = properties[key]["description"]
        option = f"{title}: {description}"
        options.append((key, option))
    # sorting: metadata on top, then parameter per key
    options = sorted(options, key=lambda k: k[0].lower())
    options = add_metadata_parameter(options)
    # restrict to search
    options = [
        key for key in options
        if (key[0].lower().find(incomplete.lower()) != -1
            or key[1].lower().find(incomplete.lower()) != -1
            )
    ]
    return options


def dataset_types(ctx, args, incomplete):
    """Prepare a list of dataset types."""
    CONTEXT.set_connection_from_args(args)
    incomplete = incomplete.lower()
    options = []
    plugins = get_task_plugins()
    for plugin_id in plugins:
        plugin = plugins[plugin_id]
        title = plugin["title"]
        description = plugin["description"].partition("\n")[0]
        option = f"{title}: {description}"
        if plugin["taskType"] == "Dataset" and (
                plugin_id.lower().find(incomplete.lower()) != -1
                or option.lower().find(incomplete.lower()) != -1):
            options.append(
                (
                    plugin_id,
                    option
                )
            )
    options = sorted(options, key=lambda k: k[1].lower())
    return options


def dataset_ids(ctx, args, incomplete):
    """Prepare a list of projectid:datasetid dataset identifier."""
    CONTEXT.set_connection_from_args(args)
    options = []
    results = list_items(item_type="dataset")
    datasets = results["results"]
    for _ in datasets:
        options.append(
            (
                _["projectId"] + r"\:" + _["id"],
                _["label"]
            )
        )
    return _finalize_completion(
        candidates=options,
        incomplete=incomplete,
        sort_by=SORT_BY_DESC
    )


def dataset_list_filter(ctx, args, incomplete):
    """Prepare a list of filter names and values for dataset list filter."""
    filter_names = [
        (
            "project",
            "Filter by project ID."
        ),
        (
            "regex",
            "Filter by regular expression on the dataset label."
        ),
        (
            "tag",
            "Filter by tag label."
        ),
        (
            "type",
            "Filter by dataset type."
        ),
    ]
    filter_regex = [
        (
            r"^Final\:",
            "Example: Dataset label starts with 'Final:'."
        ),
        (
            r"[12][0-9][0-9][0-9]-[0-9][0-9]-[0-9][0-9]",
            "Example: Dataset label contains a data-like string."
        )
    ]
    options = []
    if args[len(args) - 1] == "--filter":
        options = _finalize_completion(
            candidates=filter_names,
            incomplete=incomplete
        )
    if args[len(args) - 1] == "type":
        options = dataset_types(ctx, args, incomplete)
    if args[len(args) - 1] == "project":
        options = project_ids(ctx, args, incomplete)
    if args[len(args) - 1] == "tag":
        options = tag_labels(ctx, args, incomplete, "dataset")
    if args[len(args) - 1] == "regex":
        options = _finalize_completion(
            candidates=filter_regex,
            incomplete=incomplete
        )
    return options


def resource_ids(ctx, args, incomplete):
    """Prepare a list of projectid:resourceid resource identifier."""
    CONTEXT.set_connection_from_args(args)
    options = []
    for _ in get_all_resources():
        options.append(
            (
                _["id"]
            )
        )
    return _finalize_completion(
        candidates=options,
        incomplete=incomplete,
        sort_by=SORT_BY_DESC
    )


def scheduler_ids(ctx, args, incomplete):
    """Prepare a list of projectid:schedulerid scheduler identifier."""
    CONTEXT.set_connection_from_args(args)
    options = []
    schedulers = list_items(
        item_type="task",
        facets=[{
            "facetId": "taskType",
            "keywordIds": ["Scheduler"],
            "type": "keyword"
        }]
    )["results"]
    for _ in schedulers:
        if _["projectId"] + ":" + _["id"] in args:
            continue
        options.append(
            (
                _["projectId"] + r"\:" + _["id"],
                _["label"]
            )
        )
    return _finalize_completion(
        candidates=options,
        incomplete=incomplete,
        sort_by=SORT_BY_DESC
    )


def metric_ids(ctx, args, incomplete):
    """Prepare a list of metric identifier."""
    CONTEXT.set_connection_from_args(args)
    options = []

    data = get_prometheus_data().text
    for family in text_string_to_metric_families(data):
        options.append(
            (
                family.name,
                family.documentation
            )
        )
    return _finalize_completion(
        candidates=options,
        incomplete=incomplete,
        sort_by=SORT_BY_KEY
    )


def metric_label_filter(ctx, args, incomplete):
    """Prepare a list of label name or values."""
    # TODO: this completion does not take care of the --job option
    CONTEXT.set_connection_from_args(args)
    incomplete = incomplete.lower()
    options = []
    metric_id = args[3]
    labels = metric_get_labels(metrics_get_dict()[metric_id])
    if args[len(args) - 1] in "--filter":
        # we are in the name position
        options = labels.keys()
    if args[len(args) - 2] in "--filter":
        label_name = args[len(args) - 1]
        # we are in the value position
        options = labels[label_name]
    return _finalize_completion(
        candidates=options,
        incomplete=incomplete,
        sort_by=SORT_BY_KEY
    )


def vocabularies(ctx, args, incomplete, filter_="all"):
    """Prepare a list of vocabulary graphs for auto-completion."""
    CONTEXT.set_connection_from_args(args)
    options = []
    try:
        vocabs = get_vocabularies(filter_=filter_)
    except Exception:
        # if something went wrong, die silently
        return []
    for _ in vocabs:
        url = _["iri"]
        if url in args:
            continue
        url = _["iri"].replace(":", r"\:")
        try:
            label = _["label"]["title"]
        except (KeyError, TypeError):
            label = "Vocabulary in graph " + url
        options.append((url, label))
    return _finalize_completion(
        candidates=options,
        incomplete=incomplete,
        sort_by=SORT_BY_DESC
    )


def installed_vocabularies(ctx, args, incomplete):
    """Prepare a list of installed vocabulary graphs."""
    return vocabularies(ctx, args, incomplete, filter_="installed")


def installable_vocabularies(ctx, args, incomplete):
    """Prepare a list of installable vocabulary graphs."""
    return vocabularies(ctx, args, incomplete, filter_="installable")


def file_list(incomplete="", suffix="", description="", prefix=""):
    """Prepare a list of files with specific parameter."""
    directory = os.getcwd()
    options = []
    for file_name in os.listdir(directory):
        if os.path.isfile(file_name) \
                and file_name.endswith(suffix) \
                and file_name.startswith(prefix):
            options.append((file_name, description))
    return _finalize_completion(
        candidates=options,
        incomplete=incomplete,
        sort_by=SORT_BY_KEY
    )


def workflow_io_ids(ctx, args, incomplete):
    """Prepare a list of io workflows."""
    CONTEXT.set_connection_from_args(args)
    options = []
    for _ in get_workflows_io():
        workflow_id = _["projectId"] + r"\:" + _["id"]
        label = _["label"]
        options.append((workflow_id, label))
    return _finalize_completion(
        candidates=options,
        incomplete=incomplete,
        sort_by=SORT_BY_DESC
    )


def replay_files(ctx, args, incomplete):
    """Prepare a list of JSON replay files."""
    return file_list(
        incomplete=incomplete,
        suffix=".json",
        description="JSON query replay file"
    )


def installed_package_names(ctx, args, incomplete):
    """Prepare a list of installed packages."""
    CONTEXT.set_connection_from_args(args)
    options = []
    packages = list_packages()
    for _ in packages:
        options.append(
            (
                _["name"],
                _["version"]
            )
        )
    return _finalize_completion(
        candidates=options,
        incomplete=incomplete,
        sort_by=SORT_BY_KEY
    )


def published_package_names(ctx, args, incomplete):
    """List of plugin packages scraped from pypi.org."""
    options = []
    url = "https://pypi.org/search/?q=%22cmem-plugin-%22"
    soup = BeautifulSoup(
        requests.get(url, timeout=5).content,
        "html.parser"
    )
    packages = soup.find_all("a", class_="package-snippet")
    for package in packages:
        name = package.findChildren(
            class_="package-snippet__name"
        )[0].getText()
        if name == "cmem-plugin-base":
            continue
        description = package.findChildren(
            class_="package-snippet__description"
        )[0].getText()
        version = package.findChildren(
            class_="package-snippet__version"
        )[0].getText()
        options.append(
            (
                name,
                f"{version}: {description}"
            )
        )

    return _finalize_completion(
        candidates=options,
        incomplete=incomplete,
        sort_by=SORT_BY_KEY
    )


def python_package_files(ctx, args, incomplete):
    """Prepare a list of acceptable python package files."""
    return file_list(
        incomplete=incomplete,
        suffix=".tar.gz",
        description="Plugin Python Package file",
        prefix="cmem-plugin-"
    )


def installable_packages(ctx, args, incomplete):
    """Installable packages from files and pypi.org."""
    return python_package_files(
        ctx, args, incomplete) + published_package_names(
        ctx, args, incomplete)


def workflow_io_input_files(ctx, args, incomplete):
    """Prepare a list of acceptable workflow io input files."""
    return file_list(
        incomplete=incomplete,
        suffix=".csv",
        description="CSV Dataset resource"
    ) + file_list(
        incomplete=incomplete,
        suffix=".xml",
        description="XML Dataset resource"
    ) + file_list(
        incomplete=incomplete,
        suffix=".json",
        description="JSON Dataset resource"
    )


def workflow_io_output_files(ctx, args, incomplete):
    """Prepare a list of acceptable workflow io output files."""
    return file_list(
        incomplete=incomplete,
        suffix=".csv",
        description="CSV Dataset resource"
    ) + file_list(
        incomplete=incomplete,
        suffix=".xml",
        description="XML Dataset resource"
    ) + file_list(
        incomplete=incomplete,
        suffix=".json",
        description="JSON Dataset resource"
    ) + file_list(
        incomplete=incomplete,
        suffix=".xlsx",
        description="Excel Dataset resource"
    ) + file_list(
        incomplete=incomplete,
        suffix=".ttl",
        description="RDF file Dataset resource"
    ) + file_list(
        incomplete=incomplete,
        suffix=".nt",
        description="RDF file Dataset resource"
    )


def dataset_files(ctx, args, incomplete):
    """Prepare a list of SPARQL files."""
    return file_list(
        incomplete=incomplete,
        suffix=".csv",
        description="CSV Dataset resource"
    ) + file_list(
        incomplete=incomplete,
        suffix=".xlsx",
        description="Excel Dataset resource"
    ) + file_list(
        incomplete=incomplete,
        suffix=".xml",
        description="XML Dataset resource"
    ) + file_list(
        incomplete=incomplete,
        suffix=".json",
        description="JSON Dataset resource"
    ) + file_list(
        incomplete=incomplete,
        suffix=".ttl",
        description="RDF file Dataset resource"
    ) + file_list(
        incomplete=incomplete,
        suffix=".zip",
        description="multiCsv Dataset resource"
    ) + file_list(
        incomplete=incomplete,
        suffix=".orc",
        description="Apache ORC Dataset resource"
    )


def graph_backup_files(ctx, args, incomplete):
    """Prepare a list of workspace files."""
    return file_list(
        incomplete=incomplete,
        suffix=".graphs.zip",
        description="eccenca Corporate Memory graph backup file"
    )


def project_files(ctx, args, incomplete):
    """Prepare a list of workspace files."""
    return file_list(
        incomplete=incomplete,
        suffix=".project.zip",
        description="eccenca Corporate Memory project backup file"
    )


def ini_files(ctx, args, incomplete):
    """Prepare a list of workspace files."""
    return file_list(
        incomplete=incomplete,
        suffix=".ini",
        description="INI file"
    )


def workspace_files(ctx, args, incomplete):
    """Prepare a list of workspace files."""
    return file_list(
        incomplete=incomplete,
        suffix=".workspace.zip",
        description="eccenca Corporate Memory workspace backup file"
    )


def sparql_files(ctx, args, incomplete):
    """Prepare a list of SPARQL files."""
    return file_list(
        incomplete=incomplete,
        suffix=".sparql",
        description="SPARQL query file"
    ) + file_list(
        incomplete=incomplete,
        suffix=".rq",
        description="SPARQL query file"
    )


def triple_files(ctx, args, incomplete):
    """Prepare a list of triple files."""
    return file_list(
        incomplete=incomplete,
        suffix=".ttl",
        description="RDF Turtle file"
    ) + file_list(
        incomplete=incomplete,
        suffix=".nt",
        description="RDF NTriples file"
    )


def placeholder(ctx, args, incomplete):
    """Prepare a list of placeholder from the to-be executed queries."""
    # look if cursor is in value position of the -p option and
    # return nothing in case it is (values are not completed atm)
    if args[len(args) - 2] in ("-p", "--parameter"):
        return []
    # setup configuration
    CONTEXT.set_connection_from_args(args)
    # extract placeholder from given queries in the command line
    options = []
    for num, arg in enumerate(args):
        query = QUERY_CATALOG.get_query(arg)
        if query is not None:
            options.extend(
                list(query.get_placeholder_keys())
            )
    # look for already given parameter in the arguments and remove them from
    # the available options
    for num, arg in enumerate(args):
        if num - 1 > 0 and args[num - 1] in ("-p", "--parameter"):
            options.remove(arg)
    return _finalize_completion(
        candidates=options,
        incomplete=incomplete
    )


def remote_queries(ctx, args, incomplete):
    """Prepare a list of query URIs."""
    CONTEXT.set_connection_from_args(args)
    options = []
    for _, query in QUERY_CATALOG.get_queries().items():
        url = query.short_url.replace(":", r"\:")
        label = query.label
        options.append((url, label))
    return _finalize_completion(
        candidates=options,
        incomplete=incomplete,
        sort_by=SORT_BY_DESC
    )


def remote_queries_and_sparql_files(ctx, args, incomplete):
    """Prepare a list of named queries, query files and directories."""
    remote = remote_queries(ctx, args, incomplete)
    files = sparql_files(ctx, args, incomplete)
    return remote + files


def workflow_ids(ctx, args, incomplete):
    """Prepare a list of projectid:taskid workflow identifier."""
    CONTEXT.set_connection_from_args(args)
    workflows = list_items(item_type="workflow")["results"]
    options = []
    for _ in workflows:
        workflow = _["projectId"] + ":" + _["id"]
        label = _["label"]
        if workflow in args:
            continue
        options.append((workflow.replace(":", r"\:"), label))
    return _finalize_completion(
        candidates=options,
        incomplete=incomplete,
        sort_by=SORT_BY_DESC
    )


def marshalling_plugins(ctx, args, incomplete):
    """Prepare a list of supported workspace/project import/export plugins."""
    CONTEXT.set_connection_from_args(args)
    options = get_marshalling_plugins()
    if "description" in options[0].keys():
        return [(_["id"], _["description"]) for _ in options]
    # in case, no descriptions are available, labels are fine as well
    return [(_["id"], _["label"]) for _ in options]


def project_ids(ctx, args, incomplete):
    """Prepare a list of project IDs for auto-completion."""
    CONTEXT.set_connection_from_args(args)
    try:
        projects = get_projects()
    except Exception:
        # if something went wrong, die silently
        return []
    options = []
    for _ in projects:
        project_id = _["name"]
        label = _["metaData"]["label"]
        # do not add project if already in the command line
        if project_id in args:
            continue
        options.append((project_id, label))
    return _finalize_completion(
        candidates=options,
        incomplete=incomplete,
        sort_by=SORT_BY_DESC
    )


def graph_uris(ctx, args, incomplete, writeable=True, readonly=True):
    """Prepare a list of graphs for auto-completion."""
    CONTEXT.set_connection_from_args(args)
    try:
        graphs = get_graphs()
    except Exception:
        # if something went wrong, die silently
        return []
    options = []
    for _ in graphs:
        iri = _["iri"]
        label = _["label"]["title"]
        # do not add graph if already in the command line
        if iri in args:
            continue
        options.append((iri.replace(":", r"\:"), label))
    return _finalize_completion(
        candidates=options,
        incomplete=incomplete,
        sort_by=SORT_BY_DESC
    )


def writable_graph_uris(ctx, args, incomplete):
    """Prepare a list of writable graphs for auto-completion."""
    return graph_uris(ctx, args, incomplete, writeable=True, readonly=False)


def connections(ctx, args, incomplete):
    """Prepare a list of config connections for auto-completion."""
    # since ctx does not have an obj here, we re-create the object
    CONTEXT.set_connection_from_args(args)
    options = []
    for section in CONTEXT.config.sections():
        options.append(section)
    return _finalize_completion(
        candidates=options,
        incomplete=incomplete
    )


def graph_export_templates(ctx, args, incomplete):
    """Prepare a list of example templates for the graph export command."""
    examples = [
        (
            "{{hash}}",
            "Example: 6568a[...]00b08.ttl"
        ),
        (
            "{{iriname}}",
            "Example: https__ns_eccenca_com_data_config.ttl"
        ),
        (
            "{{date}}-{{iriname}}",
            "Example: 2021-11-29-https__ns_eccenca_com_data_config.ttl"
        ),
        (
            "{{date}}-{{connection}}-{{iriname}}",
            "Example: 2021-11-29-mycmem-https__ns_eccenca_com_data_config.ttl"
        ),
    ]
    return _finalize_completion(
        candidates=examples,
        incomplete=incomplete
    )


def project_export_templates(ctx, args, incomplete):
    """Prepare a list of example templates for the project export command."""
    examples = [
        (
            "{{id}}",
            "Example: a plain file name"
        ),
        (
            "{{date}}-{{connection}}-{{id}}.project",
            "Example: a more descriptive file name"),
        (
            "dumps/{{connection}}/{{id}}/{{date}}.project",
            "Example: a whole directory tree"
        )
    ]
    return _finalize_completion(
        candidates=examples,
        incomplete=incomplete
    )


def workspace_export_templates(ctx, args, incomplete):
    """Prepare a list of example templates for the workspace export command."""
    examples = [
        (
            "workspace",
            "Example: a plain file name"
        ),
        (
            "{{date}}-{{connection}}.workspace",
            "Example: a more descriptive file name"),
        (
            "dumps/{{connection}}/{{date}}.workspace",
            "Example: a whole directory tree"
        )
    ]
    return _finalize_completion(
        candidates=examples,
        incomplete=incomplete
    )


def query_status_filter(ctx, args, incomplete):
    """Prepare a list of filter names and values for query status filter."""
    filter_names = [
        (
            "status",
            "List only queries which have a certain status "
            "(cancelled, running, finished or error)."
        ),
        (
            "slower-than",
            "List only queries which are slower than X milliseconds."
        ),
        (
            "type",
            "List only queries of a certain query type."
        ),
        (
            "regex",
            "List only queries which query text matches a regular expression."
        ),
        (
            "trace-id",
            "List only queries which have the specified trace ID."
        ),
        (
            "user",
            "List only queries executed by the specified account (URL)."
        ),
        (
            "graph",
            "List only queries which affected a certain graph (URL)."
        ),
    ]
    filter_regex = [
        (
            r"http\:\/\/schema.org",
            "List only queries which somehow use the schema.org namespace."
        ),
        (
            r"http\:\/\/www.w3.org\/2000\/01\/rdf-schema\#",
            "List only queries which somehow use the RDF schema namespace."
        ),
        (
            r"\\\?value",
            "List only queries which are using the ?value projection variable."
        ),
        (
            "^CREATE SILENT GRAPH",
            "List only queries which start with CREATE SILENT GRAPH."
        )
    ]
    filter_status = [
        (
            "running",
            "List only queries which are currently running."
        ),
        (
            "finished",
            "List only queries which are not running anymore."
        ),
        (
            "error",
            "List only queries which were NOT successfully executed."
        ),
        (
            "cancelled",
            "List only queries which were cancelled."
        ),
        (
            "timeout",
            "List only queries which ran into a timeout."
        )
    ]
    filter_slower = [
        (
            "5",
            "List only queries which are executed slower than 5ms."
        ),
        (
            "100",
            "List only queries which are executed slower than 100ms."
        ),
        (
            "1000",
            "List only queries which are executed slower than 1000ms."
        ),
        (
            "5000",
            "List only queries which are executed slower than 5000ms."
        )
    ]
    filter_type = [
        (
            "ASK",
            "Queries, which return a boolean indicating whether a query "
            "pattern matches or not."
        ),
        (
            "SELECT",
            "Queries, which returns all, or a subset of, the variables "
            "bound in a query pattern match."
        ),
        (
            "DESCRIBE",
            "Queries, which return an RDF graph that describes the "
            "resources found."
        ),
        (
            "CONSTRUCT",
            "Queries, which return an RDF graph constructed by substituting "
            "variables in a set of triple templates."
        ),
        (
            "UPDATE", "Queries, which modify one or more graphs "
                      "(INSERT, DELETE, DROP, etc.)."
        ),
        (
            "UNKNOWN",
            "Queries of unknown type."
        )
    ]

    last_argument = args[len(args) - 1]
    options = None
    if last_argument == "--filter":
        options = _finalize_completion(
            candidates=filter_names,
            incomplete=incomplete
        )
    if last_argument == "regex":
        options = _finalize_completion(
            candidates=filter_regex,
            incomplete=incomplete
        )
    if last_argument == "status":
        options = _finalize_completion(
            candidates=filter_status,
            incomplete=incomplete
        )
    if last_argument == "slower-than":
        options = _finalize_completion(
            candidates=filter_slower,
            incomplete=incomplete,
            nat_sort=True
        )
    if last_argument == "type":
        options = _finalize_completion(
            candidates=filter_type,
            incomplete=incomplete
        )
    if last_argument == "user":
        options = query_account_iris(ctx, args, incomplete)
    if last_argument == "trace-id":
        options = query_trace_ids(ctx, args, incomplete)
    if last_argument == "graph":
        options = query_graphs(ctx, args, incomplete)

    if not options:
        raise ValueError(
            "Last argument unknown. Can not do completion."
        )
    return options


def query_account_iris(ctx, args, incomplete):
    """Prepare a list account IRIs from the query status."""
    CONTEXT.set_connection_from_args(args)
    accounts = {}
    for _ in get_query_status():
        if _["user"] in accounts:
            accounts[_["user"]] += 1
        else:
            accounts[_["user"]] = 1
    options = [
        (account.replace(":", r"\:"), f"{count} queries")
        for account, count in accounts.items()
    ]
    return _finalize_completion(
        candidates=options,
        incomplete=incomplete
    )


def query_trace_ids(ctx, args, incomplete):
    """Prepare a list trace IDs from the query status."""
    CONTEXT.set_connection_from_args(args)
    options = Counter(
        [query["traceId"] for query in get_query_status()]
    ).most_common()
    return _finalize_completion(
        candidates=options,
        incomplete=incomplete,
        sort_by=SORT_BY_DESC
    )


def query_graphs(ctx, args, incomplete):
    """Prepare a list graph URLs from the query status."""
    CONTEXT.set_connection_from_args(args)
    options = Counter()
    for query in get_query_status():
        for graph in query.get("affectedGraphs", []):
            options.update([graph.replace(":", r"\:")])
    return _finalize_completion(
        candidates=options.most_common(),
        incomplete=incomplete,
        sort_by=SORT_BY_KEY
    )


def graph_list_filter(ctx, args, incomplete):
    """Prepare a list of filter names and values for graph list filter."""
    filter_names = [
        (
            "access",
            "List only graphs which have a certain access condition "
            "(readonly or writeable)."
        ),
        (
            "imported-by",
            "List only graphs which are in the import tree of a "
            "specified graph."
        )
    ]
    filter_values_access = [
        (
            "readonly",
            "List only graphs which are NOT writable for the current user."
        ),
        (
            "writeable",
            "List only graphs which ARE writeable for the current user."
        )
    ]

    options = []
    if args[len(args) - 1] == "--filter":
        options = _finalize_completion(
            candidates=filter_names,
            incomplete=incomplete
        )
    if args[len(args) - 1] == "access":
        options = _finalize_completion(
            candidates=filter_values_access,
            incomplete=incomplete
        )
    if args[len(args) - 1] == "imported-by":
        options = graph_uris(ctx, args, incomplete)
    return options


def resource_list_filter(ctx, args, incomplete):
    """Prepare a list of filter names and values for resource list filter."""
    filter_names = [
        (
            "project",
            "Filter for file resources from a specific project."
        ),
        (
            "regex",
            "Filter for file resources with a regular expression search on "
            "the name field."
        )
    ]
    filter_values_regex = [
        (
            "csv$",
            "File resources which name ends with .csv"
        ),
        (
            "2021-10-[0-9][0-9]",
            "File resources which name has a date from 2021-10 in it"
        )
    ]

    if args[len(args) - 1] == "--filter":
        return filter_names
    if args[len(args) - 1] == "project":
        return _finalize_completion(
            candidates=project_ids(ctx, args, incomplete),
            incomplete=incomplete
        )
    if args[len(args) - 1] == "regex":
        return _finalize_completion(
            candidates=filter_values_regex,
            incomplete=incomplete
        )
    return []


def workflow_list_filter(ctx, args, incomplete):
    """Prepare a list of filter names and values for workflow list filter."""
    filter_names = [
        (
            "project",
            "Filter by project ID."
        ),
        (
            "io",
            "Filter by workflow io feature."
        ),
        (
            "regex",
            "Filter by regular expression on the workflow label."
        ),
        (
            "tag",
            "Filter by tag label."
        ),
    ]
    filter_values_io = [
        (
            "any",
            "List all workflows suitable for the io command."
        ),
        (
            "input-only",
            "List only workflows with a variable input dataset."
        ),
        (
            "output-only",
            "List only workflows with a variable output dataset."
        ),
        (
            "input-output",
            "List only workflows with a variable input and output dataset."
        ),
    ]
    filter_regex = [
        (
            r"^Final\:",
            "Example: Workflow label starts with 'Final:'."
        ),
        (
            r"[12][0-9][0-9][0-9]-[0-9][0-9]-[0-9][0-9]",
            "Example: Workflow label contains a data-like string."
        )
    ]
    options = []
    if args[len(args) - 1] == "--filter":
        options = filter_names
    if args[len(args) - 1] == "io":
        options = filter_values_io
    if args[len(args) - 1] == "project":
        options = project_ids(ctx, args, incomplete)
    if args[len(args) - 1] == "tag":
        options = tag_labels(ctx, args, incomplete, "workflow")
    if args[len(args) - 1] == "regex":
        options = filter_regex
    return _finalize_completion(
        candidates=options,
        incomplete=incomplete
    )


def tag_labels(ctx, args, incomplete, item_type):
    """Prepare a list of tag labels for a item_type."""
    datasets = list_items(item_type=item_type)
    options = []
    counts = {}
    for _dataset in datasets["results"]:
        for _tag in _dataset["tags"]:
            if _tag["label"] in counts:
                counts[_tag["label"]] += 1
            else:
                counts[_tag["label"]] = 1
    for tag, count in counts.items():
        options.append((tag, f"{count} item(s): {tag}"))
    return _finalize_completion(
        candidates=options,
        incomplete=incomplete,
        sort_by=SORT_BY_DESC,
        reverse=True
    )


def status_keys(ctx, args, incomplete):
    """Prepare a list of status keys for the admin status command."""
    options = ["all"]
    os.environ["CMEMPY_IS_CHATTY"] = "false"
    status_info = struct_to_table(get_complete_status_info())
    for _ in status_info:
        options.append(_[0])
    return _finalize_completion(
        candidates=options,
        incomplete=incomplete
    )


def user_ids(ctx, args, incomplete):
    """Prepare a list of username for admin update/delete/password command."""
    CONTEXT.set_connection_from_args(args)
    options = []
    for _ in list_users():
        options.append(
            (
                _["username"]
            )
        )
    return _finalize_completion(
        candidates=options,
        incomplete=incomplete,
        sort_by=SORT_BY_DESC
    )


def user_group_ids(ctx: Context, args, incomplete):
    """Prepare a list of group name for admin user update
     --unassign-group/--assign-group parameter"""
    CONTEXT.set_connection_from_args(args)
    if not ctx.args:
        return []
    users = get_user_by_username(username=str(ctx.args[0]))
    if not users:
        return []

    if args[len(args) - 1] == "--unassign-group":
        groups = user_groups(user_id=users[0]["id"])
    else:
        user_group_names = (
            [group["name"] for group in user_groups(user_id=users[0]["id"])]
        )
        groups = (
            [group for group in list_groups() if group["name"] not in user_group_names]
        )
    options = []
    for _ in groups:
        options.append(
            (
                _["name"]
            )
        )
    for num, arg in enumerate(args):
        if num - 1 > 0 and args[num - 1] == args[len(args) - 1]:
            with suppress(ValueError):
                options.remove(arg)
    return _finalize_completion(
        candidates=options,
        incomplete=incomplete,
        sort_by=SORT_BY_DESC
    )



"""Generate a help text and command structure graph."""
from click import Command, Group


def print_manual_graph(ctx, version):
    """Output the complete manual graph.

    Returns: None
    """
    comment = "This dataset represents the complete inline documentation " \
              "of cmemc as a graph"
    print(f"""
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>.
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#>.
@prefix void: <http://rdfs.org/ns/void#>.
@prefix xsd: <http://www.w3.org/2001/XMLSchema#>.
@prefix owl: <http://www.w3.org/2002/07/owl#>.
@prefix skos: <http://www.w3.org/2004/02/skos/core#>.
@prefix cli: <https://vocabs.eccenca.com/cli/>.
@prefix s: <http://schema.org/>.
@prefix : <https://cmemc.eccenca.dev/>.
@prefix i: <https://eccenca.com/go/cmemc>.

: a void:Dataset ;
    owl:versionInfo "{version}" ;
    rdfs:label "cmemc: Documentation Graph" ;
    rdfs:comment "{comment}" .

i: a s:SoftwareApplication ;
    rdfs:label "cmemc" ;
    s:softwareVersion "{version}" .
""")
    print_group_manual_graph_recursive(ctx.command, ctx=ctx)


def print_group_manual_graph_recursive(command_group, ctx=None, prefix=""):
    """Output documentation graph (recursive)."""
    commands = command_group.commands
    for key in commands:
        item = commands[key]
        iri = f":{prefix}{key}"
        print(f"{iri} skos:notation '{key}' .")
        if isinstance(item, Group):
            comment = item.get_short_help_str(limit=200)
            sub_group_iri = f":{prefix[:-1]}"
            if sub_group_iri == ":":
                sub_group_iri = "<https://eccenca.com/go/cmemc>"
            print(f"{iri} a cli:CommandGroup .")
            print(f"{iri} rdfs:label '{prefix}{key} Command Group' .")
            print(f"{iri} cli:subGroupOf {sub_group_iri} .")
            print(f"{iri} rdfs:comment '{comment}' .")
            print_group_manual_graph_recursive(
                item,
                ctx=ctx,
                prefix=f"{prefix}{key}-"
            )
        elif isinstance(item, Command):
            comment = item.get_short_help_str(limit=200)
            group_iri = f":{prefix[:-1]}"
            print(f"{iri} a cli:Command .")
            print(f"{iri} rdfs:label '{prefix}{key} Command' .")
            print(f"{iri} cli:group {group_iri} .")
            print(f"{iri} rdfs:comment '{comment}' .")
            for parameter in item.params:
                print_parameter_manual_graph(
                    parameter,
                    prefix=f"{prefix}{key}-"
                )
        else:
            pass


def print_parameter_manual_graph(item, prefix=""):
    """Output documentation graph for parameter."""
    iri = f":{prefix}{item.name}"
    print(f"{iri} a cli:Parameter .")
    print(f"{iri} cli:command :{prefix[:-1]}.")
    if len(item.opts) == 1:
        print(f"{iri} rdfs:label '{item.opts[0]}'.")
    else:
        print(f'{iri} rdfs:label """{max(item.opts, key=len)}""".')
        print(f'{iri} cli:shortenedOption """{min(item.opts, key=len)}""".')
    if item.default not in (None, False, [None, None]):
        print(f"{iri} cli:defaultValue '{item.default}'.")
    try:
        print(f'{iri} rdfs:comment """{item.help}""".')
    except AttributeError:
        pass
